From iris.proofmode Require Import tactics.
From lrust.typing Require Import typing.
Set Default Proof Using "Type".

Section defs.
  Context `{typeG Σ}.

  Fixpoint list_to_vec_tywflst {args} (WF : TyWfLst args) : TyWfLst (list_to_vec args) :=
    match WF with
    | tyl_wf_nil => tyl_wf_nil
    | tyl_wf_cons ty tyl _ _ => tyl_wf_cons ty (list_to_vec tyl)
    end.
  Existing Instance list_to_vec_tywflst.

  Class FnOnce (args : list type) (F : type) `{!TyWfLst args, !TyWf F} : Type :=
    { Fn_output : type;
      Fn_output_wf :> TyWf Fn_output;
      call_once : val;
      call_once_type :
        typed_val call_once (fn (λ _ : (), FP_wf ∅ (F ::: list_to_vec args) Fn_output))
    }.

  Class FnMut (args : list type) (F : type)
        `{!TyWfLst args, !TyWf F, !FnOnce args F} : Type :=
    { call_mut : val;
      call_mut_type :
        typed_val call_mut (fn (λ α, FP_wf ∅ (&uniq{α} F ::: list_to_vec args) Fn_output))
    }.

  Class Fn (args : list type) (F : type)
        `{!TyWfLst args, !TyWf F, !FnOnce args F, !FnMut args F} : Type :=
    { call : val;
      call_type :
        typed_val call (fn (λ α, FP_wf ∅ (&shr{α} F ::: list_to_vec args) Fn_output))
    }.
End defs.

Arguments Fn_output {_ _ _} _ {_ _ _}.
Arguments call_once {_ _ _} _ {_ _ _}.
Arguments call_mut {_ _ _} _ {_ _ _ _}.
Arguments call {_ _ _} _ {_ _ _ _ _}.

(* FIXME : these instances should be defined for any number of
   parameters, but the typeclass search for [Closed] is not able to
   support variable number of parameters. *)
Program Instance FnOnce_shr_bor_0 `{typeG Σ} (α : lft) (F : type)
        `{!TyWf F, !FnOnce [] F, !FnMut [] F, !Fn [] F} :
  FnOnce [] (&shr{α}F) :=
  {| Fn_output := Fn_output F;
     call_once :=
       funrec: <> ["x"] :=
         let: "x'" := !"x" in
         letalloc: "x''" <- "x'" in
         let: "f" := call F in
         letcall: "r" := "f" ["x''"]%E in
         delete [ #1; "x"];; return: ["r"]
  |}.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros ([] ϝ ret p).
  inv_vec p=>x. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_let; [by apply call_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnOnce_shr_bor_1 `{typeG Σ} (α : lft) (tyarg1 : type) (F : type) 
        `{!TyWfLst [tyarg1], !TyWf F, !FnOnce [tyarg1] F, !FnMut [tyarg1] F, !Fn [tyarg1] F} :
    FnOnce [tyarg1] (&shr{α}F) :=
  { Fn_output := Fn_output F;
    call_once :=
      funrec: <> ["x"; "a1"] :=
        let: "x'" := !"x" in
        letalloc: "x''" <- "x'" in
        let: "f" := call F in
        letcall: "r" := "f" ["x''"; "a1"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros ([] ϝ ret p).
  inv_vec p=>x a1. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_let; [by apply call_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnMut_shr_bor_0 `{typeG Σ} (α : lft) (F : type)
        `{!TyWf F, !FnOnce [] F, !FnMut [] F, !Fn [] F} :
  FnMut [] (&shr{α}F) :=
  { call_mut :=
      funrec: <> ["x"] :=
        let: "x'" := !"x" in
        let: "x''" := !"x'" in
        letalloc: "x'''" <- "x''" in
        let: "f" := call F in
        letcall: "r" := "f" ["x'''"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros (β ϝ ret p).
  inv_vec p=>x. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x'''). simpl_subst.
  iApply type_let; [by apply call_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnMut_shr_bor_1 `{typeG Σ} (α : lft) (tyarg1 : type) (F : type)
  `{!TyWfLst [tyarg1], !TyWf F, !FnOnce [tyarg1] F, !FnMut [tyarg1] F, !Fn [tyarg1] F} :
  FnMut [tyarg1] (&shr{α}F) :=
  { call_mut :=
      funrec: <> ["x"; "a1"] :=
        let: "x'" := !"x" in
        let: "x''" := !"x'" in
        letalloc: "x'''" <- "x''" in
        let: "f" := call F in
        letcall: "r" := "f" ["x'''"; "a1"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros (β ϝ ret p).
  inv_vec p=>x a1. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x'''). simpl_subst.
  iApply type_let; [by apply call_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance Fn_shr_bor_0 `{typeG Σ} (α : lft) (F : type)
        `{!TyWf F, !FnOnce [] F, !FnMut [] F, !Fn [] F} :
  Fn [] (&shr{α}F) :=
  { call :=
      funrec: <> ["x"] :=
        let: "x'" := !"x" in
        let: "x''" := !"x'" in
        letalloc: "x'''" <- "x''" in
        let: "f" := call F in
        letcall: "r" := "f" ["x'''"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros (β ϝ ret p).
  inv_vec p=>x. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x'''). simpl_subst.
  iApply type_let; [by apply call_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance Fn_shr_bor_1 `{typeG Σ} (α : lft) (tyarg1 : type) (F : type)
        `{!TyWfLst [tyarg1], !TyWf F, !FnOnce [tyarg1] F, !FnMut [tyarg1] F, !Fn [tyarg1] F} :
  Fn [tyarg1] (&shr{α}F) :=
  { call :=
      funrec: <> ["x";"a1"] :=
        let: "x'" := !"x" in
        let: "x''" := !"x'" in
        letalloc: "x'''" <- "x''" in
        let: "f" := call F in
        letcall: "r" := "f" ["x'''";"a1"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros (β ϝ ret p).
  inv_vec p=>x a1. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x'''). simpl_subst.
  iApply type_let; [by apply call_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnOnce_uniq_bor_0 `{typeG Σ} (α : lft) (F : type)
        `{!TyWf F, !FnOnce [] F, !FnMut [] F} :
    FnOnce [] (&uniq{α}F) :=
  { Fn_output := Fn_output F;
    call_once :=
      funrec: <> ["x"] :=
        let: "x'" := !"x" in
        letalloc: "x''" <- "x'" in
        let: "f" := call_mut F in
        letcall: "r" := "f" ["x''"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros ([] ϝ ret p).
  inv_vec p=>x. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_let; [by apply call_mut_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnOnce_uniq_bor_1 `{typeG Σ} (α : lft) (tyarg1 : type) (F : type)
        `{!TyWfLst [tyarg1], !TyWf F, !FnOnce [tyarg1] F, !FnMut [tyarg1] F} :
  FnOnce [tyarg1] (&uniq{α}F) :=
  { Fn_output := Fn_output F;
    call_once :=
      funrec: <> ["x";"a1"] :=
        let: "x'" := !"x" in
        letalloc: "x''" <- "x'" in
        let: "f" := call_mut F in
        letcall: "r" := "f" ["x''";"a1"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros ([] ϝ ret p).
  inv_vec p=>x a1. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_let; [by apply call_mut_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnMut_uniq_bor_0 `{typeG Σ} (α : lft) (F : type)
        `{!TyWf F, !FnOnce [] F, !FnMut [] F} :
  FnMut [] (&uniq{α}F) :=
  { call_mut :=
      funrec: <> ["x"] :=
        let: "x'" := !"x" in
        let: "x''" := !"x'" in
        letalloc: "x'''" <- "x''" in
        let: "f" := call_mut F in
        letcall: "r" := "f" ["x'''"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros (β ret ϝ p).
  inv_vec p=>x. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_deref_uniq_uniq; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x'''). simpl_subst.
  iApply type_let; [by apply call_mut_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.

Program Instance FnMut_uniq_bor_1 `{typeG Σ} (α : lft) (tyarg1 : type) (F : type)
        `{!TyWfLst [tyarg1], !TyWf F, !FnOnce [tyarg1] F, !FnMut [tyarg1] F} :
  FnMut [tyarg1] (&uniq{α}F) :=
  { call_mut :=
      funrec: <> ["x";"a1"] :=
        let: "x'" := !"x" in
        let: "x''" := !"x'" in
        letalloc: "x'''" <- "x''" in
        let: "f" := call_mut F in
        letcall: "r" := "f" ["x'''";"a1"]%E in
        delete [ #1; "x"];; return: ["r"]
  }.
Next Obligation.
  intros. iApply type_fn; [solve_typing..|]. iIntros "/= !#". iIntros (β ret ϝ p).
  inv_vec p=>x a1. simpl_subst.
  iApply type_deref; [solve_typing..|]. iIntros (x'). simpl_subst.
  iApply type_deref_uniq_uniq; [solve_typing..|]. iIntros (x''). simpl_subst.
  iApply type_letalloc_1; [solve_typing..|]. iIntros (x'''). simpl_subst.
  iApply type_let; [by apply call_mut_type|solve_typing|]. iIntros (fc). simpl_subst.
  iApply type_letcall; [solve_typing..|]. iIntros (r). simpl_subst.
  iApply type_delete; [solve_typing..|].
  iApply type_jump; solve_typing.
Qed.
